#![allow(dead_code, unused_variables)]

use std::str::FromStr;

/// Parse the string `s` as a coordinate pair, like `"400x600"` or `"1.0,0.5"`.
///
/// Specifically, `s` should have the form <left><sep><right>, where <sep> is
/// the character given by the `separator` argument, and <left> and <right> are both
/// strings that can be parsed by `T::from_str`.
///
/// If `s` has the proper form, return `Some<(x, y)>`. If it doesn't parse
/// correctly, return `None`.
fn parse_pair<T: FromStr>(s: &str, separator: char) -> Option<(T, T)> {
    match s.find(separator) {
        None => None,
        Some(index) => {
            match (T::from_str(&s[..index]), T::from_str(&s[index + 1..])) {
                (Ok(l), Ok(r)) => Some((l, r)),
                _ => None
            }
        }
    }
}

#[test]
fn test_parse_pair() {
    assert_eq!(parse_pair::<i32>("",        ','), None);
    assert_eq!(parse_pair::<i32>("10,",     ','), None);
    assert_eq!(parse_pair::<i32>(",10",     ','), None);
    assert_eq!(parse_pair::<i32>("10,20",   ','), Some((10, 20)));
    assert_eq!(parse_pair::<i32>("10,20xy", ','), None);
    assert_eq!(parse_pair::<f64>("0.5x",    'x'), None);
    assert_eq!(parse_pair::<f64>("0.5x1.5", 'x'), Some((0.5, 1.5)));
}

/// Return the point on the complex plane corresponding to a given pixel in the
/// bitmap.
///
/// `bornes` is a pair giving the large and haute of the bitmap. `pixel` is a
/// pair indicating a particular pixel in that bitmap. The `super_g` and
/// `infer_d` parameters are points on the complex plane designating the
/// area our bitmap covers.
fn pixel_to_point(bornes: (usize, usize),
                  pixel:  (usize, usize),
                  super_g: (f64, f64),
                  infer_d: (f64, f64))
    -> (f64, f64)
{
    // It might be nicer to find the position of the *middle* of the pixel,
    // instead of its upper left corner, but this is easier to write tests for.
    let (large, haute) = (infer_d.0 - super_g.0,
                          super_g.1 - infer_d.1);
    (super_g.0 + pixel.0 as f64 * large / bornes.0 as f64,
     super_g.1 - pixel.1 as f64 * haute / bornes.1 as f64)
}

#[test]
fn test_pixel_to_point() {
    assert_eq!(pixel_to_point((100, 100), (25, 75),
                              (-1.0, 1.0), (1.0, -1.0)),
               (-0.5, -0.5));
}

extern crate num;
use num::Complex;

/// Try to determine whether the complex number `c` is in the Mandelbrot set.
///
/// A number `c` is in the set if, starting with zero, repeatedly squaring and
/// adding `c` never causes the number to leave the circle of radius 2 centered
/// on the origin; the number instead orbits near the origin forever. (If the
/// number does leave the circle, it eventually flies away to infinity.)
///
/// If after `limit` iterations our number has still not left the circle, return
/// `None`; this is as close as we come to knowing that `c` is in the set.
///
/// If the number does leave the circle before we give up, return `Some(i)`, where
/// `i` is the number of iterations it took.
fn escapes(c: Complex<f64>, limit: u32) -> Option<u32> {
    let mut z = Complex { re: 0.0, im: 0.0 };
    for i in 0..limit {
        z = z*z + c;
        if z.norm_sqr() > 4.0 {
            return Some(i);
        }
    }

    return None;
}

/// Render a rectangle of the Mandelbrot set into a buffer of pixels.
///
/// The `bornes` argument gives the large and haute of the buffer `pixels`,
/// which holds one grayscale pixel per byte. The `super_g` and `infer_d`
/// arguments specify points on the complex plane corresponding to the upper
/// left and lower right corners of the pixel buffer.
fn render(pixels: &mut [u8], bornes: (usize, usize),
          super_g: (f64, f64), infer_d: (f64, f64))
{
    assert!(pixels.len() == bornes.0 * bornes.1);

    for r in 0 .. bornes.1 {
        for c in 0 .. bornes.0 {
            let point = pixel_to_point(bornes, (c, r),
                                       super_g, infer_d);
            pixels[r * bornes.0 + c] =
                match escapes(Complex { re: point.0, im: point.1 }, 255) {
                    None => 0,
                    Some(count) => 255 - count as u8
                };
        }
    }
}

extern crate image;

use std::fs::File;
use std::io::Result;
use image::png::PNGEncoder;
use image::ColorType;

/// Write the buffer `pixels`, whose dimensions are given by `bornes`, to the
/// file named `nomfic`.
fn write_bitmap(nomfic: &str, pixels: &[u8], bornes: (usize, usize))
    -> Result<()>
{
    let output = try!(File::create(nomfic));

    let encodeur = PNGEncoder::new(output);
    try!(encodeur.encode(&pixels[..],
                        bornes.0 as u32, bornes.1 as u32,
                        ColorType::Gray(8)));

    Ok(())
}

use std::io::Write;

fn main() {
    let args : Vec<String> = std::env::args().collect();

    if args.len() != 5 {
        writeln!(std::io::stderr(),
                 "Usage: mandelbrot NOMFIC PIXELS COINSUPG COININFD")
            .unwrap();
        writeln!(std::io::stderr(),
                 "Exemple: {} mandel.png 1000x750 -1.20,0.35 -1,0.20",
                 args[0])
            .unwrap();
        std::process::exit(1);
    }

    let bornes = parse_pair(&args[2], 'x')
        .expect("error parsing image dimensions");
    let super_g = parse_pair(&args[3], ',')
        .expect("error parsing upper left corner point");
    let infer_d = parse_pair(&args[4], ',')
        .expect("error parsing lower right corner point");

    let mut pixels = vec![0; bornes.0 * bornes.1];
    render(&mut pixels[..], bornes, super_g, infer_d);

    write_bitmap(&args[1], &pixels[..], bornes).expect("error writing PNG file");
}
